/* Jedina razlika u odnosu na originalni program je sto se umesto nadovezivanja putanja koristi menjanje tekuceg direktorijuma.
   Kada se naidje na neki direktorijum D, onda se tekuci direktorijum postavlja na D. Pri citanju fajlova iz D nece biti problema
   jer su svi fajlovi vidljivi iz tekuceg direktorijuma. Kada se direktorijum obradi, vrsi se povratak u naddirektorijum. */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include <dirent.h>
#include <string.h>

/* Struktura sadrzi broj fajlova za svaki tip fajla. */
typedef struct _Statistics {
  int regular_count;
  int directory_count;
  int fifo_count;
  int block_count;
  int character_count;
  int link_count;
  int socket_count;
  int total_count;
} Statistics;

void traverse (char *file_name, Statistics *stats);
void error_fatal (const char *message);
void error_fatal_file (const char *message, const char *file_name);

/* Program stampa procenat svakog od tipova fajlova medju ukupnim brojem faljova u direktorijumu cije se
   ime navodi kao argument komandne linije. */
int main(int argc, char **argv){
  assert (argc == 2);
  Statistics statistics = {0, 0, 0, 0, 0, 0, 0, 0};

  traverse (argv[1], &statistics);

  if (statistics.total_count != 0){
    printf ("regular files           : %6.2f%%\n", (100.0 * statistics.regular_count) / statistics.total_count);
    printf ("directories             : %6.2f%%\n", (100.0 * statistics.directory_count) / statistics.total_count);
    printf ("character special files : %6.2f%%\n", (100.0 * statistics.character_count) / statistics.total_count);
    printf ("block special files     : %6.2f%%\n", (100.0 * statistics.block_count) / statistics.total_count);
    printf ("fifo files              : %6.2f%%\n", (100.0 * statistics.fifo_count) / statistics.total_count);
    printf ("socket files            : %6.2f%%\n", (100.0 * statistics.socket_count) / statistics.total_count);
    printf ("symbolic links          : %6.2f%%\n", (100.0 * statistics.link_count) / statistics.total_count);
  }

  return 0;
}

/* Rekurzivna funkcija koja sakuplja informacije o tipovima fajlova u sturkturu tipa Statistics. */
void traverse (char *file_name, Statistics *statistics){
  struct stat file_info;

  /* Sakupljaju se informacije o fajlu pomocu lstat funkcije. */
  if (lstat (file_name, &file_info) < 0)
    error_fatal_file ("lstat error: ", file_name);

  /* Bez obzira na tip fajla, ukupan broj fajlova se povecava. */
  statistics->total_count++;

  /* U zavisnosti od tipa, povecava se odgovarajuci brojac. */
  if (S_ISREG (file_info.st_mode))
    statistics->regular_count++;
  else if (S_ISFIFO (file_info.st_mode))
    statistics->fifo_count++;
  else if (S_ISCHR (file_info.st_mode))
    statistics->character_count++;
  else if (S_ISBLK (file_info.st_mode))
    statistics->block_count++;
  else if (S_ISSOCK (file_info.st_mode))
    statistics->socket_count++;
  else if (S_ISLNK (file_info.st_mode))
    statistics->link_count++;
  else if (S_ISDIR (file_info.st_mode)){
    DIR *directory;
    struct dirent *entry;

    /* Ako se naislo na direktorijum, mora se proci i kroz sve fajlove koji se u njemu nalaze. */

    /* Otvaranje direktorijuma. */
    if ( (directory = opendir (file_name)) < 0) 
      error_fatal_file ("opendir error", file_name);

    /* Promena tekuceg direktorijuma. */
    if (chdir (file_name) < 0)
      error_fatal_file ("chdir error", file_name);

    /* Citanje direktorijumskih odrednica. */
    while ( (entry = readdir (directory)) != NULL){

      /* Ako se naidje na odrednicu . ili .. preskace se (u suprotnom bi postojala beskonacna petlja.  */
      if (strcmp (entry->d_name, ".") == 0 || strcmp (entry->d_name, "..") == 0)
	continue;

      /* Rekurzivni poziv funkije za fajl u direktorijumu. */
      traverse (entry->d_name, statistics);
    }

    statistics->directory_count++;
    
    /* Vracanje u naddirektorijum. */
    if (chdir ("..") < 0)
      error_fatal_file ("chdir error for ..", file_name);

    if (closedir (directory) < 0)
      error_fatal_file ("closedir error", file_name);
  }
  

}

/* Funkcija za obradu greske. */
void error_fatal (const char *message){
  perror (message);
  exit (EXIT_FAILURE);
}

/* Funkcija za obradu greske. */
void error_fatal_file (const char *message, const char *file_name){
  perror (message);
  fprintf (stderr, "%s", file_name);
  exit (EXIT_FAILURE);
}

